from .classification import GaussianProcessClassifier
from .regression import GaussianProcessRegressor
from .edr import EffectiveDimensionalityReduction

__all__ = ['GaussianProcessClassifier',
           'GaussianProcessRegressor',
           'EffectiveDimensionalityReduction']
